/* Функции Значимости объектов восприятия

Значимость всегда определяется в контексте всех предшествующих условий,
т.е. специфична для активностей деревьев автоматизмов и понимания.

При каждом вызове consciousnessElementary определяется текущий объект наибольшой значимости в воспринимаемом -
в функции определения текущей Цели getMentalPurpose().
*/

package psychic

import (
	word_sensor "BOT/brain/words_sensor"
	"BOT/lib"
)

/////////////////////////////////////////////////
/* Фиксация значимости объекта ОБъективного восприятия и действия.
Значимость - величина от -N до N, приобретаемая объектов внимания в данной ситуации
- берется из результата пробных действий и связывается со всеми компонентами воспринимаемого в этих условиях
Тот объект, применение которого привело к данной значимости.
Вызывается в момент ответных действий оператора в период ожидания.

т.е. к эффекту от текущего действия привязываются два объекта:
образ текущего Стимула curActiveActions в ответ на совершенное действие LastAutomatizmWeiting.ActionsImageID
*/
func setImportance(effect int){

	/* для Стимула curActiveActions - структура действий оператора при активации дерева автоматизмов
	 */
	setImportanceForObject(curActiveActions,effect)

	// для Ответного действия Beast
	if LastAutomatizmWeiting !=nil {

		//act:=ActionsImageArr[LastAutomatizmWeiting.ActionsImageID]
		act,ok:=ReadeActionsImageArr(LastAutomatizmWeiting.ActionsImageID)
		if ok {
			setImportanceForObject(act, effect)
		}
	}
}
/* эффект для составляющих общего образа в 3 раза меньше,
потому что при усреднении значимости integralObjectImportance() для всех ProblemID частота составляющих будет больше,
чем более сложного образа.
 */
func setImportanceForObject(obj *ActionsImage,effect int){
	if detectedActiveLastProblemNodID==0{
		return
	}
	if obj==nil{
		return
	}
	// целостный образ действий ID ActionsImage
	createNewlastImportanceID(0, detectedActiveLastProblemNodID, 1, obj.ID,effect, true)

	// эффект для составляющих общего образа
	sE:=secondaryEffect(effect)

	// ID несловестного действия ActionsImage.ActID[n]
	if len(obj.ActID) > 0{
		for i := 0; i < len(obj.ActID); i++ {
			createNewlastImportanceID(0, detectedActiveLastProblemNodID, 3, obj.ActID[i],sE, true)
		}
	}

	// ID Verbal - при активации дерева автоматизмов
	if curActiveVerbalID >0 {
		createNewlastImportanceID(0, detectedActiveLastProblemNodID, 4, curActiveVerbalID, sE, true)
	}

	// ID отдельной фразы Verbal.PhraseID[n]
	if len(obj.PhraseID) >0{
		for i := 0; i < len(obj.PhraseID); i++ {
			createNewlastImportanceID(0, detectedActiveLastProblemNodID, 5, obj.PhraseID[i],sE, true)
			// по каждому слову фразы
			wRr:=word_sensor.GetWordArrFromPhraseID(obj.PhraseID[i])
			for j := 0; j < len(wRr); j++ {
				createNewlastImportanceID(0, detectedActiveLastProblemNodID, 6,wRr[j],sE, true)
			}
		}
	}

	// ID тон сообщения с Пульта  Verbal.ToneID
	if obj.ToneID>0{
		createNewlastImportanceID(0, detectedActiveLastProblemNodID, 7, obj.ToneID, sE, true)
	}

	// ID настроение оператора  Verbal.MoodID
	if obj.MoodID>0{
		createNewlastImportanceID(0, detectedActiveLastProblemNodID, 8, obj.MoodID, sE, true)
	}
}
// Из исходного Эффекта вычислисть вторичный эффект для состоавляющих образа
func secondaryEffect(effect int)int{
	sE:=int(effect/3)
	if sE==0{
		if effect>0{ sE=1 }
		if effect<0{ sE=-1 }
	}
return sE
}
////////////////////////////////////////////////////////

/* Фиксация значимости объекта СУБъективного восприятия.

заполняется из afterWaitingPeriod(
*/
func setImportanceMental(effect int){

}
////////////////////////////////////////////////////



/////////////////////////////////////////////////////


/*Найти наиболее значащие объекты в данных условиях
при этом помещая их в importanceFromID=make(map[int]*importance)
 */
func getGreatestImportance(curActions *ActionsImage)[]extremImportance{
	var importanceObjectArr []extremImportance
	if curActions == nil{
		return nil
	}
	var id,v = 0,0
	// целостный образ действий
	id,v=getObjectsImportanceValue(1,curActions.ID, detectedActiveLastProblemNodID)
	if id>0{
		// в качестве объекта подстаовляем curActions.ID потому, что это ActionsImageArr.ID, а не importanceConditinArr.ID
		eID,_:=createNewlastImportanceID(0, detectedActiveLastProblemNodID, 1, curActions.ID, v, true)
		importanceObjectArr =append(importanceObjectArr,extremImportance{eID,curActions.ID,1,v})
	}

	// ID несловестного действия ActionsImage.ActID[n]
	if len(curActions.ActID) > 0{
		max:=0
		actID:=0
		for i := 0; i < len(curActions.ActID); i++ {
			id,v=getObjectsImportanceValue(3,curActions.ActID[i], detectedActiveLastProblemNodID)
			if id>0{
				if lib.Abs(max) > lib.Abs(v) {
					max = v
					actID=curActions.ActID[i]
				}
			}
		}
		if actID>0{
			eID,_:=createNewlastImportanceID(0, detectedActiveLastProblemNodID, 3, actID, max, true)
			importanceObjectArr =append(importanceObjectArr,extremImportance{eID,actID,3,max})
		}
	}

	// ID Verbal - при активации дерева автоматизмов
	if curActiveVerbalID >0 {
		id,v=getObjectsImportanceValue(4,curActiveVerbalID, detectedActiveLastProblemNodID)
		if id>0{
			eID,_:=createNewlastImportanceID(0, detectedActiveLastProblemNodID, 4, curActiveVerbalID, v, true)
			importanceObjectArr =append(importanceObjectArr,extremImportance{eID,curActiveVerbalID,4,v})
		}
	}

	//ID отдельной фразы Verbal.PhraseID[n]
	if len(curActions.PhraseID) >0 {
		max:=0
		actID:=0
		for i := 0; i < len(curActions.PhraseID); i++ {
			id,v=getObjectsImportanceValue(5,curActions.PhraseID[i], detectedActiveLastProblemNodID)
			if id>0{
				if lib.Abs(max) > lib.Abs(v){
					max = v
					actID=curActions.PhraseID[i]
				}
			}

			//ID отдельного слова  из Verbal.PhraseID[n]
			max2:=0
			actID2:=0

			wRr:=word_sensor.GetWordArrFromPhraseID(curActions.PhraseID[i])
			for j := 0; j < len(wRr); j++ {
				id,v=getObjectsImportanceValue(6,wRr[j], detectedActiveLastProblemNodID)
				if id>0{
					if lib.Abs(max2) > lib.Abs(v){
						max2 = v
						actID2=wRr[j]
					}
				}
			}
			if actID2>0{
				eID,_:=createNewlastImportanceID(0, detectedActiveLastProblemNodID, 6, actID2, max2, true)
				importanceObjectArr =append(importanceObjectArr,extremImportance{eID,actID2,6,max2})
			}

		}
		if actID>0{
			eID,_:=createNewlastImportanceID(0, detectedActiveLastProblemNodID, 5, actID, max, true)
			importanceObjectArr =append(importanceObjectArr,extremImportance{eID,actID,5,max})
		}
	}

	// ID тон сообщения с Пульта  Verbal.ToneID
	if curActions.ToneID>0{
		id,v=getObjectsImportanceValue(7,curActions.ToneID, detectedActiveLastProblemNodID)
		if id>0{
			eID,_:=createNewlastImportanceID(0, detectedActiveLastProblemNodID, 7, curActions.ToneID, v, true)
			importanceObjectArr =append(importanceObjectArr,extremImportance{eID,curActions.ToneID,7,v})
		}
	}

	// ID настроение оператора  Verbal.MoodID
	if curActions.MoodID>0{
		id,v=getObjectsImportanceValue(8,curActions.MoodID, detectedActiveLastProblemNodID)
		if id>0{
			eID,_:=createNewlastImportanceID(0, detectedActiveLastProblemNodID, 8, curActions.MoodID, v, true)
			importanceObjectArr =append(importanceObjectArr,extremImportance{eID,curActions.MoodID,8,v})
		}
	}

	return importanceObjectArr
}
/////////////////////////////////////////////////

/* выбрать один, самый актуальный (максимально ннегативный или позитипный) объект
из curImportanceJbjectArr []extremImportance
Возвращает индекс массива curImportanceObjectArr или -1
 */
func getTopAttentionObject(eij []extremImportance)int{
	if eij==nil{
		return -1
	}
	// тупо выбираем самый значимый
	max:=0
	maxIndex:=-1
	var maxE *extremImportance

	for i := 0; i < len(eij); i++ {
		if lib.Abs(max) < lib.Abs(eij[i].extremVal){
			max = eij[i].extremVal
			maxIndex=i
			maxE=&eij[i]
		}
	}
	if maxE!=nil{
		/*Сoздание нового элемента модели понимания в условиях его выявления
Раз выявлен экстремальный объект внимания, то для него пишется эпизодическая память без правил,
		только для Модели понимания.
		Для этого maxE *extremImportance преобразуем в тип reflexes.ActiveCurTriggerStimulsID
		 */
		saveNewEpisodic(maxE.extremObjID,maxE.kind,maxE.objID,1)

		/* тут де раньше заполнялись категориии для categorization.go
		// updateCategiryModel(epmID,index)
		Старая поддержка категорий убрана, можно посмотреть в categorization.go и categorization_functions.go
		в архиве 2024_01_14_try_episodic_tree_rekease_FINAL_OF_3_VERSION
		ВНИМАНИЕ! любые взаимосвязи теперь находятся по дереву эпиз.памяти.
		т.к. поддержка категорий пока еще не применена нигде,
		то ОСТАВЛЯЕТСЯ НА БУДУЩЕЕ при необходимости.
		Тогда и будут написаны все необходимые функции.
		Пока что удаляется набор категорий и запись файла /memory_psy/category_model.txt
		который не нужен т.к. поиск по дереву эпиз.памяти достаточно быстрый.
		 */

		return maxIndex
	}

	return 0
}
/////////////////////////////////////////////////

/* выбрать один, самый позитивный объект из curImportanceJbjectArr []extremImportance
Возвращает индекс массива curImportanceObjectArr или -1
*/

/////////////////////////////////////////////////


/* Средняя значимость данного образа для всех случаях его восприятия
(т.е. независимо от ProblemID)
Это - общее представление о значимости образа.
TODO пока не используется, но может применяьтся для обращения внимания на такой образ.
 */
func integralObjectImportance(Type int,ObjectID int)int{
	var iVal=0
	for _, v := range importanceFromID {
		if v==nil{continue}
		if v.Type!=Type || v.ObjectID!=ObjectID{
			continue
		}
		iVal+=v.Value
	}
return iVal
}
/////////////////////////////////////////////////////





////////////////////////////////////////////////////
/* выделить наиболее значимое в мыслях в текущем цикле - привлечение внимания к собственным мыслям.
При каждом СУБъективном вызове consciousnessElementary
 */
type extremImportanceMental struct {
	objID int//  объект значимости
	kind int // тип объекта
	extremVal int// экстремальная значимость
}
var curImportanceObjectMentalArr []extremImportanceMental // сюда складываются текущие цели внимания к наиболее важному
func getGreatestImportanceMental(){

// TODO - объект внимания к собственным мыслям
}
/////////////////////////////////////////////////









